;;; org-comment.el --- Comment-type link syntax for Org  -*- lexical-binding: t; -*-

;; Copyright (C) 2015  Eric Abrahamsen

;; Author: Eric Abrahamsen <eric@ericabrahamsen.net>

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; Provides a new link type for Org that allows you to create comments
;; on arbitrary chunks of text.  The link prefix is "comment:".

;; Add comments with `org-comment-add-comment'.  Following the link
;; will display the text of the comment in a pop-up buffer.  The
;; buffer is in special-mode, hit "q" to dismiss it.

;; Call `org-comment-display-comments' to see all comments in a buffer.

;; See the `org-comment-[backend]-export-style' options for ways to
;; format comments in export.

;; Todo:

;; 1. Better export customization options.

;; 2. What does the ODT comment XML look like?

;; 3. More functions in the display comment buffer: copy as
;; kill... what else?

;; 4. More functions in the comments list buffer, to display, pop to,
;; delete, and edit comment text.

;; 5. Is it possible to have multi-line filled tabular list items?
;; Long comments are not very useful if you can't see the whole thing.

;; 5. Allow multiple comment list buffers attached to different Org
;; buffers.

;; 6. Maybe a minor mode for ease of manipulating comments?

;;; With thanks to John Kitchen for getting the ball rolling:
;;; http://kitchingroup.cheme.cmu.edu/blog/2015/04/24/Commenting-in-org-files/

;;; Code:

(require 'org)
(require 'cl-lib)

(org-add-link-type
 "comment"
 #'org-comment-display-comment
 #'org-comment-export-comment)

(defgroup org-comment nil
  "Comment link type for Org."
  :tag "Org Comments"
  :group 'org)

(defcustom org-comment-display-buffer "*Org Comment*"
  "Name of the buffer for temporary display of comment text."
  :group 'org-comment
  :type 'string)

(defcustom org-comment-list-buffer "*Org Comment List*"
  "Name of the buffer for displaying all comments in an Org buffer/subtree."
  :group 'org-comment
  :type 'string)

(defcustom org-comment-html-export-style 'tooltip
  "The HTML export style for Org comments, as a symbol. Currently
only supports tooltip."
  :group 'org-comment
  :type '(choice
	  (const :tag "tooltip" tooltip)))

(defcustom org-comment-latex-export-style 'marginpar
  "The LaTeX export style for Org comments, as a
symbol. Currently supports marginpar, todonote, and footnote."
  :group 'org-comment
  :type '(choice
	  (const :tag "marginpar" marginpar)
	  (const :tag "todonote" todonote)
	  (const :tag "footnote" footnote)))

(defcustom org-comment-odt-export-style 'comment
  "The ODT export style for Org comments, as a symbol.  Currently
only supports comment."
  :group 'org-comment
  :type '(choice
	  (const :tag "comment" comment)))

;; The purpose of having this buffer-local is defeated by the fact
;; that we only have one *Org Comment List* buffer!
(defvar-local org-comment-comments-source nil
  "Buffer/marker pair pointing to the source of comments for a
  given comment-list buffer.")

(defun org-comment-export-comment (path desc format)
  (cl-case format
    ('html
     (cl-case org-comment-html-export-style
       ('tooltip
	(format "<font color=\"red\"><abbr title=\"%s\" color=\"red\">COMMENT</abbr></font> %s" path (or desc "")))))
    ('latex
     (cl-case org-comment-latex-export-style
       ('marginpar
	(format "%s\\marginpar{%s}" (or desc "") path))
       ('todonote
	(format "%s\\todo{%s}" (or desc "") path))
       ('footnote
	(format "%s\\footnote{%s}" (or desc "") path))))
    (t ;; Figure out ODT export.
     "")))

(defun org-comment-display-comment (linkstring)
  (when linkstring
    (with-current-buffer
	(get-buffer-create org-comment-display-buffer)
      (let ((inhibit-read-only t))
	(erase-buffer)
	(insert linkstring)))
    (display-buffer-below-selected
     (get-buffer-create org-comment-display-buffer)
     '(nil (window-height . fit-window-to-buffer)))
    (select-window (get-buffer-window org-comment-display-buffer) t)
    (special-mode)))

;;;###autoload
(defun org-comment-add-comment ()
  (interactive)
  (if (use-region-p)
      (let ((selected-text
	     (buffer-substring (region-beginning) (region-end))))
        (setf (buffer-substring (region-beginning) (region-end))
              (format "[[comment:%s][%s]]"
                      (read-string "Comment: ") selected-text)))
  (insert (format "[[comment:%s]]" (read-string "Comment: ")))))

;;;###autoload
(defun org-comment-display-comments (arg)
  "Display all comments in the current buffer (or, with a prefix
arg, in the current subtree) in a tabulated list form."
  (interactive "P")
  (let ((buf (current-buffer))
	(marker (when arg
		  (save-excursion
		    (org-back-to-heading t)
		    (point-marker)))))
   (switch-to-buffer-other-window
    (get-buffer-create org-comment-list-buffer))
   (org-comment-list-mode)
   (setq org-comment-comments-source (cons buf marker))
   (org-comment-refresh-list)))

(defun org-comment-collect-links (&optional arg)
  "Do the work of finding all the comments in the current buffer
or subtree."
  (when org-comment-comments-source
    (with-current-buffer (car org-comment-comments-source)
      (save-restriction
	(widen)
	(let* ((marker (cdr org-comment-comments-source))
	       (beg (or marker (point-min)))
	       (end (if marker
			(save-excursion
			  (goto-char marker)
			  (outline-next-heading)
			  (point))
		      (point-max)))
	       links)
	  (goto-char beg)
	  (while (re-search-forward org-bracket-link-regexp end t)
	    (let ((path (match-string-no-properties 1))
		  (text (match-string-no-properties 3)))
	      (when (string-match-p "\\`comment:" path)
		;; Don't collect comments with no
		(setq path
		      (org-link-unescape
		       (replace-regexp-in-string
			"\n+" " "
			(replace-regexp-in-string "\\`comment:" "" path))))
		(setq text (if text
			       (org-link-unescape
				(replace-regexp-in-string "\n+" " " text))
			     "[no text]"))
		;; The format required by tabular list mode.
		(push (list (point) (vector text path)) links))))
	  (when links
	    (reverse links)))))))

(defun org-comment-refresh-list ()
  (let ((links (org-comment-collect-links))
	(max-width 0))
    (if links
	(progn
	  (dolist (l links)
	    (setq max-width
		  (max max-width
		       (string-width (aref (cadr l) 0)))))
	  (setq tabulated-list-entries links
		tabulated-list-format
		(vector `("Text" ,(min max-width 40) t) '("Comment" 40 t)))
	  (tabulated-list-init-header)
	  (tabulated-list-print))
      (message "No comments found"))))

(define-derived-mode org-comment-list-mode
    tabulated-list-mode "Org Comments"
  "Mode for viewing Org comments as a tabular list."
  (setq tabulated-list-sort-key nil)
  (add-hook 'tabulated-list-revert-hook
	    #'org-comment-refresh-list))

(provide 'org-comment)
;;; org-comment.el ends here
